// gcc -o usb_jjy -D_DEBUG -D_DEBUG_CONS -I. -lftdi1 usb_jjy.c

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <syslog.h>
#include <fcntl.h>		/* open() */
#include <sys/types.h>		/* open() */
#include <sys/stat.h>		/* open() */
#include <sys/time.h>
#include <ftdi.h>

#define OFFSET 5000

void reset();
void pos_marker(int *sec);
void out0(int *sec);
void out1(int *sec);
void waitNextSec(int *sec);
void ftdi_fatal (struct ftdi_context *ftdi, char *str);

struct ftdi_context *ftdi;

static void handler(int sig)
{
  unsigned char data=0;
  if (ftdi_write_data(ftdi, &data, 1) < 0)
  {
      openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
      syslog(LOG_ERR, "write failed at handler().");
      closelog();
#ifdef _DEBUG_CONS
      fprintf(stderr,"write failed for 0x%x: %s\n",
      data, ftdi_get_error_string(ftdi));
#endif
  }
  openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
  syslog(LOG_INFO, "USB JJY stopped.");
  closelog();
#ifdef _DEBUG_CONS
  fprintf(stderr,"interrupted.\n");
#endif
  ftdi_usb_close(ftdi);
  ftdi_free(ftdi);
  exit (0);
}

int main(int argc, char **argv)
{
  time_t timer;
  struct tm *t_st;

  int pa1,pa2,jy_t;
  int num, *sec;

  FILE *fp;

  // daemonize
  if (fork() != 0) exit(0);
  if (setsid() < 0)
  {
    fprintf(stderr, "setsid()\n");
    exit(1);
  }
  if (fork() != 0) exit(0);
  chdir("/"); close(0); close(1); close(2);

  if (signal(SIGHUP,  handler) == SIG_ERR) return 1;
  if (signal(SIGINT,  handler) == SIG_ERR) return 1;
  if (signal(SIGTERM, handler) == SIG_ERR) return 1;

  openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
  syslog(LOG_NOTICE, "USB JJY started.");
  closelog();

  fp = fopen("/var/run/jjyd.pid","w");
  fprintf(fp,"%d",getpid());
  fclose(fp);

  reset();

  while(1) {
    do {
      time(&timer);
      t_st = localtime(&timer);
      usleep(1000);
    } while ((int)t_st->tm_sec != 0);

    pa1 = 0; pa2 = 0; num = 0; sec = &num;

#ifdef _DEBUG_CONS
    printf("年  : %d\n",t_st->tm_year);
    printf("日数: %d\n",t_st->tm_yday);
    printf("月  : %d\n",t_st->tm_mon+1);
    printf("日  : %d\n",t_st->tm_mday);
    printf("曜日: %d\n",t_st->tm_wday);
    printf("時  : %d\n",t_st->tm_hour);
    printf("分  : %d\n",t_st->tm_min);
    printf("秒  : %d\n",t_st->tm_sec);
#endif

    pos_marker(sec);

    jy_t = t_st->tm_min;
    if (jy_t - 40 >= 0) { out1(sec); jy_t -= 40; pa2 = !pa2; } else out0(sec);
    if (jy_t - 20 >= 0) { out1(sec); jy_t -= 20; pa2 = !pa2; } else out0(sec);
    if (jy_t - 10 >= 0) { out1(sec); jy_t -= 10; pa2 = !pa2; } else out0(sec);
    out0(sec);
    if (jy_t -  8 >= 0) { out1(sec); jy_t -=  8; pa2 = !pa2; } else out0(sec);
    if (jy_t -  4 >= 0) { out1(sec); jy_t -=  4; pa2 = !pa2; } else out0(sec);
    if (jy_t -  2 >= 0) { out1(sec); jy_t -=  2; pa2 = !pa2; } else out0(sec);
    if (jy_t -  1 >= 0) { out1(sec); jy_t -=  1; pa2 = !pa2; } else out0(sec);
    pos_marker(sec);

    jy_t = t_st->tm_hour;
    out0(sec);
    out0(sec);
    if (jy_t - 20 >= 0) { out1(sec); jy_t -= 20; pa1 = !pa1; } else out0(sec);
    if (jy_t - 10 >= 0) { out1(sec); jy_t -= 10; pa1 = !pa1; } else out0(sec);
    out0(sec);
    if (jy_t -  8 >= 0) { out1(sec); jy_t -=  8; pa1 = !pa1; } else out0(sec);
    if (jy_t -  4 >= 0) { out1(sec); jy_t -=  4; pa1 = !pa1; } else out0(sec);
    if (jy_t -  2 >= 0) { out1(sec); jy_t -=  2; pa1 = !pa1; } else out0(sec);
    if (jy_t -  1 >= 0) { out1(sec); jy_t -=  1; pa1 = !pa1; } else out0(sec);
    pos_marker(sec);

    jy_t = t_st->tm_yday + 1;
    out0(sec);
    out0(sec);
    if (jy_t - 200 >= 0) { out1(sec); jy_t -= 200; } else out0(sec);
    if (jy_t - 100 >= 0) { out1(sec); jy_t -= 100; } else out0(sec);
    out0(sec);
    if (jy_t -  80 >= 0) { out1(sec); jy_t -=  80; } else out0(sec);
    if (jy_t -  40 >= 0) { out1(sec); jy_t -=  40; } else out0(sec);
    if (jy_t -  20 >= 0) { out1(sec); jy_t -=  20; } else out0(sec);
    if (jy_t -  10 >= 0) { out1(sec); jy_t -=  10; } else out0(sec);
    pos_marker(sec);

    if (jy_t -   8 >= 0) { out1(sec); jy_t -=   8; } else out0(sec);
    if (jy_t -   4 >= 0) { out1(sec); jy_t -=   4; } else out0(sec);
    if (jy_t -   2 >= 0) { out1(sec); jy_t -=   2; } else out0(sec);
    if (jy_t -   1 >= 0) { out1(sec); jy_t -=   1; } else out0(sec);
    out0(sec);
    out0(sec);
    if (pa1) out1(sec); else out0(sec);
    if (pa2) out1(sec); else out0(sec);
    out0(sec);
    pos_marker(sec);

    jy_t = t_st->tm_year % 100;
    out0(sec);
    if (jy_t - 80 >= 0) { out1(sec); jy_t -= 80; } else out0(sec);
    if (jy_t - 40 >= 0) { out1(sec); jy_t -= 40; } else out0(sec);
    if (jy_t - 20 >= 0) { out1(sec); jy_t -= 20; } else out0(sec);
    if (jy_t - 10 >= 0) { out1(sec); jy_t -= 10; } else out0(sec);
    if (jy_t -  8 >= 0) { out1(sec); jy_t -=  8; } else out0(sec);
    if (jy_t -  4 >= 0) { out1(sec); jy_t -=  4; } else out0(sec);
    if (jy_t -  2 >= 0) { out1(sec); jy_t -=  2; } else out0(sec);
    if (jy_t -  1 >= 0) { out1(sec); jy_t -=  1; } else out0(sec);
    pos_marker(sec);
    
    jy_t = t_st->tm_wday;
    if (jy_t - 4 >= 0) { out1(sec); jy_t -= 4; } else out0(sec);
    if (jy_t - 2 >= 0) { out1(sec); jy_t -= 2; } else out0(sec);
    if (jy_t - 1 >= 0) { out1(sec); jy_t -= 1; } else out0(sec);
    out0(sec);
    out0(sec);
    out0(sec);
    out0(sec);
    out0(sec);
    out0(sec);
    pos_marker(sec);
  }

  ftdi_usb_close(ftdi);
  ftdi_free(ftdi);
  exit (0);
}

double get_time_sec(){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec + (double)tv.tv_usec*1e-6;
}

void reset()
{
    unsigned char data;
    int i;

    if ((ftdi = ftdi_new()) == 0)
    {
        openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
        syslog(LOG_ERR, "ftdi_bew failed at reset().");
        closelog();
#ifdef _DEBUG_CONS
        fprintf(stderr, "ftdi_bew failed\n");
#endif
        return;
    }

    if (ftdi_usb_open(ftdi, 0x0403, 0x6001) < 0)
        ftdi_fatal (ftdi, "Can't open ftdi device");

    if (ftdi_set_bitmode(ftdi, 0xFF, BITMODE_BITBANG) < 0)
        ftdi_fatal (ftdi, "Can't enable bitbang");

    data = 0;
    if (ftdi_write_data(ftdi, &data, 1) < 0)
    {
        openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
        syslog(LOG_ERR, "write failed at reset().");
        closelog();
#ifdef _DEBUG_CONS
        fprintf(stderr,"write failed for 0x%x: %s\n",
        data, ftdi_get_error_string(ftdi));
#endif
    }
}

void pos_marker(int *sec)
{
  unsigned char data;
  double start,end;
  long ns;

#ifdef _DEBUG
  openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
  syslog(LOG_ERR, "pos_marker() start");
  closelog();
#endif

#ifdef _DEBUG_CONS
  start = get_time_sec();
#endif
  data = (unsigned char)0xff;
  if (ftdi_write_data(ftdi, &data, 1) < 0)
  {
      openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
      syslog(LOG_ERR, "write failed at pos_marker().");
      closelog();
#ifdef _DEBUG_CONS
      fprintf(stderr,"write failed for 0x%x: %s\n",
      data, ftdi_get_error_string(ftdi));
#endif

      ftdi_usb_close(ftdi);
      ftdi_free(ftdi);
      exit (1);
  }
  usleep(200000 - OFFSET);
#ifdef _DEBUG_CONS
  end = get_time_sec();
#endif
  waitNextSec(sec);
#ifdef _DEBUG_CONS
  printf("%dsec: mark: %.9f sec\n", *sec,  end - start);
#endif

#ifdef _DEBUG
  openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
  syslog(LOG_ERR, "pos_marker() end");
  closelog();
#endif
}

void out0(int *sec)
{
  unsigned char data;
  double start,end;
  long ns;

#ifdef _DEBUG
  openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
  syslog(LOG_ERR, "out0() start");
  closelog();
#endif

#ifdef _DEBUG_CONS
  start = get_time_sec();
#endif
  data = (unsigned char)0xff;
  if (ftdi_write_data(ftdi, &data, 1) < 0)
  {
      openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
      syslog(LOG_ERR, "write failed at out0().");
      closelog();
#ifdef _DEBUG_CONS
      fprintf(stderr,"write failed for 0x%x: %s\n",
      data, ftdi_get_error_string(ftdi));
#endif

      ftdi_usb_close(ftdi);
      ftdi_free(ftdi);
      exit (1);
  }
  usleep(800000 - OFFSET);
#ifdef _DEBUG_CONS
  end = get_time_sec();
#endif
  waitNextSec(sec);
#ifdef _DEBUG_CONS
  printf("%dsec: out 0: %.9f sec\n", *sec, end - start);
#endif
#ifdef _DEBUG
  openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
  syslog(LOG_ERR, "out0() end");
  closelog();
#endif
}

void out1(int *sec)
{
  unsigned char data;
  double start,end;
  long ns;

#ifdef _DEBUG
  openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
  syslog(LOG_ERR, "out1() start");
  closelog();
#endif

#ifdef _DEBUG_CONS
  start = get_time_sec();
#endif
  data = (unsigned char)0xff;
  if (ftdi_write_data(ftdi, &data, 1) < 0)
  {
      openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
      syslog(LOG_ERR, "write failed at out0().");
      closelog();
#ifdef _DEBUG_CONS
      fprintf(stderr,"write failed for 0x%x: %s\n",
      data, ftdi_get_error_string(ftdi));
#endif

      ftdi_usb_close(ftdi);
      ftdi_free(ftdi);
      exit (1);
  }
  usleep(500000 - OFFSET);
#ifdef _DEBUG_CONS
  end = get_time_sec();
#endif
  waitNextSec(sec);
#ifdef _DEBUG_CONS
  printf("%dsec: out 1: %.9f sec\n", *sec,  end - start);
#endif
#ifdef _DEBUG
  openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
  syslog(LOG_ERR, "out1() end");
  closelog();
#endif
}

void waitNextSec(int *sec)
{
  unsigned char data;
  time_t timer;
  struct tm *t_st;
  char buf[64];

#ifdef _DEBUG
  sprintf(buf, "waitNextSec: %d",*sec);
  openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
  syslog(LOG_ERR, buf);
  closelog();
#endif

  data = (unsigned char)0;
  if (ftdi_write_data(ftdi, &data, 1) < 0)
  {
      openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
      syslog(LOG_ERR, "write failed at waitNextSec().");
      closelog();
#ifdef _DEBUG_CONS
      fprintf(stderr,"write failed for 0x%x: %s\n",
      data, ftdi_get_error_string(ftdi));
#endif

      ftdi_usb_close(ftdi);
      ftdi_free(ftdi);
      exit (1);
  }
  do {
    usleep(OFFSET / 10);
    time(&timer);
    t_st = localtime(&timer);
  } while ((int)t_st->tm_sec == *sec);
  *sec = (int)t_st->tm_sec;

#ifdef _DEBUG
  sprintf(buf, "waitNextSec finish: %d",*sec);
  openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
  syslog(LOG_ERR, buf);
  closelog();
#endif
}

void ftdi_fatal (struct ftdi_context *ftdi, char *str)
{
#ifdef _DEBUG
    openlog("usb_jjy", LOG_CONS | LOG_PID, LOG_USER);
    syslog(LOG_ERR, "fatal error.");
    closelog();
#endif
#ifdef _DEBUG_CONS
    fprintf (stderr, "%s: %s\n",
             str, ftdi_get_error_string (ftdi));
#endif
    ftdi_free(ftdi);
    exit(1);
}
